/* tga2df3_2.c
 * this file reimplements 'tga2df3.c' to generate colour-separated
 * ('r','g','b') outputs, as specified by Stephen McAvoy.
 * the original code was posted by "jhu" on 4.7.2012 in the povray
 * (NNTP) newsgroup "povray.unix".
 * TODO this code (like jhu's) takes it "on trust" that all tga files
 * are 24-bit, uncompressed, and same XY.  need code to read and
 * check tga headers.
 * note  do *not* install or run this program with privileges.
 */

/* cf strdup(3) */
#define _POSIX_C_SOURCE 200809L

#include <arpa/inet.h>
#include <assert.h>
#include <dirent.h>
#include <errno.h>
#include <inttypes.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/statvfs.h>
#include <unistd.h>

/* input stuff.  edit image size limits to suit */
static char **todos;
static size_t ntodo;

#define CHRDOT  '.'
#define TGAEXTN  ".tga"

static size_t imgx, imgy;

#define IMGXMAX  4096
#define IMGYMAX  4096

/* output stuff, three colours */
static FILE *df3[3];
static char *dfn[3];

/* longest directory or file name */
#define BUFSZ  1024

/* libc errors */
static char const *self_;

#define CALLFAIL                                          \
do {                                                      \
  int const e = errno;                                    \
  fprintf(stderr, "%s: error %s\n", self_, strerror(e));  \
  exit(EXIT_FAILURE);                                     \
} while (0)

/* endianmess */
typedef union Word Word;
union Word {
  uint8_t raw[6];
  uint16_t v[3];
};

/* mkTodos()
 * -------------------------------------------------------------------------- *
 * create sorted array of pointers to .tga filenames.  everything allocated.
 * return number of files to do.  does not return on error.
 * the assert() will fail in (the unlikely) case a tga gets added or deleted
 * between the two passes over directory.
 */
static int cmpFname(void const *p1, void const *p2) {
  return strcmp(*(char * const *)p1, *(char * const *)p2);
}

static size_t mkTodos(char const * const dnam, char const * const arg) {
  DIR *dp = opendir(dnam);
  if (!dp) CALLFAIL;

  size_t const alen = strlen(arg);
  size_t n = 0;
  struct dirent *p;
  char *x;
  while ((p = readdir(dp)))
    if (!(x = strrchr(p->d_name, CHRDOT)))
      continue;
    else if (strcmp(x, TGAEXTN))
      continue;
    else if (strncmp(p->d_name, arg, alen))
      continue;
    else
      n += 1;

  if (!n  ||  UINT16_MAX < n) {
    (void)closedir(dp);
    return 0;
  }

  todos = calloc(n + 1, sizeof(char *));
  if (!todos) CALLFAIL;

  rewinddir(dp);
  while ((p = readdir(dp)))
    if (!(x = strrchr(p->d_name, CHRDOT)))
      continue;
    else if (strcmp(x, TGAEXTN))
      continue;
    else if (strncmp(p->d_name, arg, alen))
      continue;
    else if (!(todos[ntodo] = strdup(p->d_name)))
      CALLFAIL;
    else
      ntodo += 1;

  if (closedir(dp)) CALLFAIL;

  assert(ntodo == n);
  qsort(todos, ntodo, sizeof(char *), cmpFname);
  return ntodo;
}

/* prepDf3s()
 * -------------------------------------------------------------------------- *
 * get image size.  for each colour check df3 doesn't exist, create it,
 * write the (identical) header.  return true if all went well.
 */
static bool getImageXY(void) {
  FILE *fp = fopen(todos[0], "rb");
  if (!fp)
    CALLFAIL;
  else if (fseek(fp, 12, SEEK_SET))
    CALLFAIL;

  bool f = false;
  Word u;
  if (2 != fread(&u.raw, sizeof(uint8_t), 2, fp))
    CALLFAIL;
  int tmp = u.raw[1] * 256 + u.raw[0];
  if (1 > tmp  ||  IMGXMAX < tmp)
    goto label_fail;
  imgx = (size_t)tmp;

  if (2 != fread(&u.raw, sizeof(uint8_t), 2, fp))
    CALLFAIL;
  tmp = u.raw[1] * 256 + u.raw[0];
  if (1 > tmp  ||  IMGYMAX < tmp)
    goto label_fail;
  imgy = (size_t)tmp;

  f = true;

label_fail:
  (void)fclose(fp);
  return f;
}

bool haveDiskSpace(size_t const x, size_t const y, size_t const z) {
  errno = 0;
  struct statvfs vfb;
  int const rc = statvfs(".", &vfb);
  if (-1 == rc)
    CALLFAIL;
  else if (!vfb.f_bavail)
    return false;
  /* plus 8k for overheads */
  uint64_t const n = ((uint64_t)(x * y * z + 8192) / vfb.f_bsize) * 3;
  return n < vfb.f_bavail;
}

static bool prepDf3s(char const * const arg) {
  if (!getImageXY())
    return false;
  else if (!haveDiskSpace(imgx, imgy, ntodo))
    return false;

  char const rgb[] = { 'r', 'g', 'b' };
  Word u;
  u.v[0] = htons((uint16_t)imgx);
  u.v[1] = htons((uint16_t)imgy);
  u.v[2] = htons((uint16_t)ntodo);

  for (size_t i = 0; 3 != i; ++i) {
    char buf[BUFSZ];
    (void)snprintf(buf, BUFSZ, "%s_%c.df3", arg, rgb[i]);
    if (!(dfn[i] = strdup(buf)))
      CALLFAIL;
    else if ((df3[i] = fopen(buf, "rb"))) {  /* exists? */
      (void)fclose(df3[i]);
      df3[i] = NULL;
      return false;
    }
    else if (!(df3[i] = fopen(buf, "wb")))
      CALLFAIL;
    else if (6 != fwrite(&u.raw, sizeof(uint8_t), 6, df3[i]))
      CALLFAIL;
  }

  return true;
}

/* convert()
 * -------------------------------------------------------------------------- *
 * for each file to do read the image pixel-by-pixel and add corresponding
 * values to df3s.  does not return on error.
 * TODO check each image same size?
 */
static int convert(void) {
  for (size_t i = 0; todos[i]; ++i) {
    FILE *fp = fopen(todos[i], "rb");
    if (!fp)
      CALLFAIL;
    else if (fseek(fp, 18, SEEK_SET))
      CALLFAIL;
    printf("adding file '%s'\n", todos[i]);
    for (size_t j = 0; (imgx * imgy) != j; ++j) {
      uint8_t pixel[3];
      if (3 != fread(pixel, sizeof(uint8_t), 3, fp))
        CALLFAIL;
      for (size_t k = 0; 3 != k; ++k)
        if (1 != fwrite(&pixel[k], sizeof(uint8_t), 1, df3[k]))
          CALLFAIL;
    }
    (void)fclose(fp);
  }

  for (size_t i = 0; 3 != i; ++i)
    if (!fclose(df3[i]))
      df3[i] = NULL;

  return EXIT_SUCCESS;
}

/* ------------------------------------------------------------------------- */

static void tidyup(void) {
  for (size_t i = 0; 3 != i; ++i)
    if (df3[i]) {
      (void)fclose(df3[i]);
      (void)unlink(dfn[i]);
    }
  for (size_t i = 0; 3 != i; ++i)
    free(dfn[i]);
  if (ntodo)
    for (size_t i = 0; ntodo != i; ++i)
      free(todos[i]);
  free(todos);
}

int main(int argc, char *argv[]) {
  self_ = argv[0];

  if (2 != argc) {
    fprintf(stderr, "usage: %s basename\n", self_);
    exit(EXIT_FAILURE);
  }
  else if (atexit(tidyup))
    CALLFAIL;

  char wd[BUFSZ] = {'\0'};
  if (!getcwd(wd, BUFSZ))
    CALLFAIL;
  else if (!mkTodos(wd, argv[1])) {
    fprintf(stderr, "error no file matches '%s*.tga' (or too many).\n",
            argv[1]);
    return 1;
  }
  else if (!prepDf3s(argv[1])) {
    fprintf(stderr, "preparing DF3s for '%s' failed.\n", argv[1]);
    return 1;
  }

  return convert();
}

/* ------------------------------------------------------------- *
 *  the content above is covered by the GNU Public License V3+.  *
 *  copyright (c) 2017 jr <creature.eternal@gmail.com>.          *
 *  all rights reserved.                                         *
 * ------------------------------------------------------------- */
